'''
A tokeniser.
'''

from abc import ABC


__all__ = [
    'Tokeniser',
]


class Tokeniser(ABC):
    '''
    Top-level abstract class representing all tokenisers.
    '''

    def tokenise(
        self,
        text: str,
    ) -> list[str]:
        '''
        Tokenise a text into a list of tokens.

        :param text: The text to tokenise.
        :return: The list of tokens.
        '''
        raise NotImplementedError()

    def tokenise_indices(
        self,
        text: str,
    ) -> list[tuple[int, int]]:
        '''
        Tokenise a text and return the indices of the tokens.
        A list of integer pair tuples ``[(i, j)]`` is returned such that
        ``text[i:j]`` is a token.

        :param text: The text to tokenise.
        :return: The list of tuple pairs containing integers specifying the
            locations of the tokens in the text.
        '''
        raise NotImplementedError()

    def detokenise(
        self,
        tokens: list[str],
    ) -> str:
        '''
        Detokenise the list of tokens back into a whole text.
        The default behaviour is to just join all the tokens with spaces in between.

        :param tokens: The tokenised text.
        :return: The text.
        '''
        return ' '.join(tokens)



'''
Korpus Malti tokeniser.
'''

import re
from malti.tokeniser.regex_tokeniser import RegexTokeniser


__all__ = [
    'KMTokeniser',
]


class KMTokeniser(RegexTokeniser):
    '''
    The tokeniser used by the MLRS Korpus Malti corpus.
    Adapted from https://github.com/UMSpeech/MASRI/blob/main/masri/tokenise/tokenise.py
    Even though the linked repository does not have an MIT license, we have permission from the
    owner, albertgatt, to include it in this MIT licensed project.
    '''

    NUMERIC_DATE = r'\d{1,2}[-/]\d{1,2}[-/]\d{2,4}|\d{2,4}[-/]\d{1,2}[-/]\d{1,2}'
    '''Capture dates expressed numerically e.g. 10/10/2010'''

    DECIMAL = r'\d+[.,/]\d+'
    '''Captures decimal numbers e.g. 10.1'''

    NUMBER = r'\d+'
    '''Captures whole numbers e.g. 10'''

    DEF_ARTICLE = r'\w{0,5}?[dtlrnsxzcżċ]-'
    '''Captures definite articles e.g. għall- or l-'''

    DEF_NUMERAL = r'-i[dtlrnsxzcżċ]'
    '''Captures definite numerals e.g. -il (as in ħdax-il)'''

    PROCLITIC_PREP = r'^\w[\'’]$'
    '''Captures proclitic prepositions e.g. l' '''

    WORD = r'\w+[`\']?|\S'
    '''Captures words e.g. kelb'''

    END_PUNCTUATION = r'\?|\.|,|\!|;|:|…|"|\'|\.\.\.\''
    '''Captures end-of-sentence punctuation marks e.g. .'''

    ABBREV_PREFIX = r'sant[\'’]|(a\.?m|p\.?m|onor|sra|nru|dott|kap|mons|dr|prof)\.?'
    '''Captures abbreviations e.g. Sant' (as in Sant'Anna)'''

    def __init__(
        self,
    ) -> None:
        '''
        Constructor.
        '''
        super().__init__(
            '|'.join([
                self.NUMERIC_DATE,
                self.DECIMAL,
                self.NUMBER,
                self.DEF_ARTICLE,
                self.DEF_NUMERAL,
                self.PROCLITIC_PREP,
                self.WORD,
                self.END_PUNCTUATION,
            ]),
            re.UNICODE | re.MULTILINE | re.DOTALL | re.IGNORECASE
        )

        # Tokens that match this regex should not have a space AFTER them.
        self._detok_no_space_after_re = re.compile(
            f'{self.DEF_ARTICLE}|{self.PROCLITIC_PREP}|{self.ABBREV_PREFIX}',
            re.IGNORECASE
        )

        # Tokens that match this regex should not have a space BEFORE them.
        self._detok_no_space_before_re = re.compile(
            f'{self.END_PUNCTUATION}|{self.DEF_NUMERAL}',
            re.IGNORECASE
        )

    def detokenise(
        self,
        tokens: list[str],
    ) -> str:
        '''
        Detokenise the list of tokens back into a whole text.

        :param tokens: The tokenised text.
        :return: The text.
        '''
        tokens_with_spaces: list[str] = []

        for token in tokens:
            if self._detok_no_space_after_re.match(token):
                tokens_with_spaces.append(token)
            elif self._detok_no_space_before_re.match(token):
                if tokens_with_spaces and tokens_with_spaces[-1] == ' ':
                    tokens_with_spaces.pop() # Remove the previously added space (if there is one).
                    tokens_with_spaces.append(token)
                    tokens_with_spaces.append(' ')
                # Code that this was taken from did not have an `else` for the previous `if`.
            else:
                tokens_with_spaces.append(token)
                tokens_with_spaces.append(' ')

        text = ''.join(tokens_with_spaces)
        return text.strip()



'''
Sentence splitters for Maltese text.
'''

from malti.sent_splitter.sent_splitter import SentSplitter
from malti.sent_splitter.km_sent_splitter.km_sent_splitter import KMSentSplitter


def split(
    text: str,
) -> list[str]:
    '''
    Default sentence splitter.
    In this version, ``KMSentenceSplitter`` is used.

    :param text: The text to split.
    :return: The list of sentences.
    '''
    return KMSentSplitter().split(text)



'''
Korpus Malti sentence splitter.
'''

import os
import sentence_splitter
from malti.sent_splitter.sent_splitter import SentSplitter


__all__ = [
    'KMSentSplitter',
]


class KMSentSplitter(SentSplitter):
    '''
    The sentence splitter used by the MLRS Korpus Malti corpus.
    '''

    def __init__(
        self,
    ) -> None:
        '''
        Constructor.
        '''
        super().__init__()
        self._spltter = sentence_splitter.SentenceSplitter(
            language='it',
            non_breaking_prefix_file=os.path.join(
                os.path.dirname(__file__),
                'mt_non_breaking_prefixes.txt',
            ),
        )

    def split(
        self,
        text: str,
    ) -> list[str]:
        '''
        Split a text into a list of sentences.

        :param text: The text to split.
        :return: The list of sentences.
        '''
        return self._spltter.split(text)



'''
Tokeniser class for representing tokenisers that work with regular expressions.
'''

import re
from malti.tokeniser.tokeniser import Tokeniser


__all__ = [
    'RegexTokeniser',
]


class RegexTokeniser(Tokeniser):
    '''
    Tokenise a text by using regular expressions that match tokens.
    '''

    def __init__(
        self,
        pattern: str,
        flags: re.RegexFlag = (
            re.UNICODE | re.MULTILINE | re.DOTALL | re.IGNORECASE
        ),
    ) -> None:
        '''
        Create a regular expression tokeniser from a regular expression.

        :param pattern: A string regular expression which will be compiled into
            a regular expression ``re`` object.
        :param flags: Regular expression flags to use from the ``re`` module.
        '''
        self._regex = re.compile(pattern, flags)

    def tokenise(
        self,
        text: str,
    ) -> list[str]:
        '''
        Tokenise a text into a list of tokens.

        :param text: The text to tokenise.
        :return: The list of tokens.
        '''
        return self._regex.findall(text)

    def tokenise_indices(
        self,
        text: str,
    ) -> list[tuple[int, int]]:
        '''
        Tokenise a text and return the indices of the tokens.
        A list of integer pair tuples ``[(i, j)]`` is returned such that
        ``text[i:j]`` is a token.

        :param text: The text to tokenise.
        :return: The list of tuple pairs containing integers specifying the
            locations of the tokens in the text.
        '''
        return [
            m.span()
            for m in self._regex.finditer(text)
        ]



'''
Tokenisers for Maltese text.
'''

from malti.tokeniser.tokeniser import Tokeniser
from malti.tokeniser.regex_tokeniser import RegexTokeniser
from malti.tokeniser.km_tokeniser.km_tokeniser import KMTokeniser


def tokenise(
    text: str,
) -> list[str]:
    '''
    Default tokeniser.
    In this version, ``KMTokeniser`` is used.

    :param text: The text to tokenise.
    :return: The list of tokens.
    '''
    return KMTokeniser().tokenise(text)



'''
A sentence splitter.
'''

from abc import ABC


__all__ = [
    'SentSplitter',
]


class SentSplitter(ABC):
    '''
    Top-level abstract class representing all sentence splitters.
    '''

    def split(
        self,
        text: str,
    ) -> list[str]:
        '''
        Split a text into a list of sentences.

        :param text: The text to split.
        :return: The list of sentences.
        '''
        raise NotImplementedError()
